@@ -23,6 +23,7 @@ import uproot.interpretation.grouped
 import uproot.language.python
 import uproot.source.chunk
 from uproot._util import no_filter, unset
+from uproot.behaviors.TBranch import _regularize_array_cache
 
 
 def iterate(
@@ -615,7 +616,7 @@ class HasFields(Mapping):
         entry_start=None,
         entry_stop=None,
         decompression_executor=None,  # TODO: Not implemented yet
-        array_cache="inherit",  # TODO: Not implemented yet
+        array_cache="inherit",
         library="ak",  # TODO: Not implemented yet
         backend="cpu",
         interpreter="cpu",
@@ -659,7 +660,7 @@ class HasFields(Mapping):
                 is used. (Not implemented yet.)
             array_cache ("inherit", None, MutableMapping, or memory size): Cache of arrays;
                 if "inherit", use the file's cache; if None, do not use a cache;
-                if a memory size, create a new cache of this size. (Not implemented yet.)
+                if a memory size, create a new cache of this size.
             library (str or :doc:`uproot.interpretation.library.Library`): The library
                 that is used to represent arrays. Options are ``"np"`` for NumPy,
                 ``"ak"`` for Awkward Array, and ``"pd"`` for Pandas. (Not implemented yet.)
@@ -725,6 +726,8 @@ class HasFields(Mapping):
             [c.num_entries for c in clusters[start_cluster_idx:stop_cluster_idx]]
         )
 
+        array_cache = _regularize_array_cache(array_cache, self.ntuple._file)
+
         form, field_path = self.to_akform(
             filter_name=filter_name,
             filter_typename=filter_typename,
@@ -747,17 +750,22 @@ class HasFields(Mapping):
                 clusters_datas,
                 start_cluster_idx,
                 stop_cluster_idx,
-                pad_missing_element=True,
             )
 
         for key in target_cols:
-            if "column" in key and "union" not in key:
+            if "column" in key:
                 key_nr = int(key.split("-")[1])
+                # Find how many elements should be padded at the beginning
+                n_padding = self.ntuple.column_records[key_nr].first_element_index
+                n_padding -= cluster_starts[start_cluster_idx]
+                n_padding = max(n_padding, 0)
                 if interpreter == "cpu":
-                    content = self.ntuple.read_col_pages(
+                    content = self.ntuple.read_cluster_range(
                         key_nr,
-                        range(start_cluster_idx, stop_cluster_idx),
-                        pad_missing_element=True,
+                        start_cluster_idx,
+                        stop_cluster_idx,
+                        missing_element_padding=n_padding,
+                        array_cache=array_cache,
                     )
                 elif interpreter == "gpu" and backend == "cuda":
                     content = content_dict[key_nr]
@@ -775,7 +783,6 @@ class HasFields(Mapping):
             form,
             cluster_num_entries,
             container_dict,
-            allow_noncanonical_form=True,
             backend="cuda" if interpreter == "gpu" and backend == "cuda" else "cpu",
         )[entry_start:entry_stop]
 
@@ -1710,7 +1717,7 @@ def _num_entries_for(ntuple, akform, target_num_bytes, entry_start, entry_stop):
 
     total_bytes = 0
     for key in target_cols:
-        if "column" in key and "union" not in key:
+        if "column" in key:
             key_nr = int(key.split("-")[1])
             for cluster in range(start_cluster_idx, stop_cluster_idx):
                 pages = ntuple.page_link_list[cluster][key_nr].pages
@@ -1792,21 +1799,14 @@ def _fill_container_dict(container_dict, content, key, dtype_byte):
             indices = _cupy_insert(indices, missing, -1)
         container_dict[f"{key}-index"] = indices
     elif dtype_byte == uproot.const.rntuple_col_type_to_num_dict["switch"]:
-        kindex, tags = uproot.models.RNTuple._split_switch_bits(content)
+        tags = content["tag"].astype(numpy.int8)
+        kindex = content["index"]
         # Find invalid variants and adjust buffers accordingly
-        invalid = numpy.flatnonzero(tags == -1)
-        if len(invalid) > 0:
-            kindex = numpy.delete(kindex, invalid)
-            tags = numpy.delete(tags, invalid)
-            invalid -= numpy.arange(len(invalid))
-            optional_index = numpy.insert(
-                numpy.arange(len(kindex), dtype=numpy.int64), invalid, -1
-            )
-        else:
-            optional_index = numpy.arange(len(kindex), dtype=numpy.int64)
-        container_dict[f"{key}-index"] = library.array(optional_index)
-        container_dict[f"{key}-union-index"] = library.array(kindex)
-        container_dict[f"{key}-union-tags"] = library.array(tags)
+        invalid = numpy.flatnonzero(tags == 0)
+        kindex[invalid] = 0  # Might not be necessary, but safer
+        container_dict[f"{key}-index"] = library.array(kindex)
+        container_dict[f"{key}-tags"] = library.array(tags)
+        container_dict["nones-index"] = library.array([-1], dtype=numpy.int64)
     else:
         # don't distinguish data and offsets
         container_dict[f"{key}-data"] = content
