@@ -16,11 +16,42 @@ from __future__ import annotations
 import ast
 import numbers
 import re
+import warnings
 
 import numpy
 
 import uproot
 
+_registered_interpretations = set()
+
+
+def register_interpretation(cls):
+    """
+    Register a custom interpretation class.
+
+    Args:
+        cls (type): A subclass of :doc:`uproot.interpretation.custom.CustomInterpretation`.
+
+    This method registers a custom interpretation class to be used in
+    :doc:`uproot.interpretation.identify.interpretation_of`.
+    """
+    if cls in _registered_interpretations:
+        warnings.warn(f"Overwriting existing custom interpretation {cls}", stacklevel=2)
+    _registered_interpretations.add(cls)
+
+
+def unregister_interpretation(cls):
+    """
+    Unregister a custom interpretation class.
+
+    Args:
+        cls (type): A subclass of :doc:`uproot.interpretation.custom.CustomInterpretation`.
+
+    This method removes a custom interpretation class from the registry.
+    """
+    if cls in _registered_interpretations:
+        _registered_interpretations.remove(cls)
+
 
 def _normalize_ftype(fType):
     if fType is not None and uproot.const.kOffsetL < fType < uproot.const.kOffsetP:
@@ -107,7 +138,7 @@ def _leaf_to_dtype(leaf, getdims):
 
 _title_has_dims = re.compile(r"^([^\[\]]*)(\[[^\[\]]+\])+")
 _item_dim_pattern = re.compile(r"\[([1-9][0-9]*)\]")
-_item_any_pattern = re.compile(r"\[(.*)\]")
+_item_any_pattern = re.compile(r"(\[.*\])")
 
 
 def _from_leaves_one(leaf, title):
@@ -312,6 +343,22 @@ def interpretation_of(branch, context, simplify=True):
     If no interpretation can be found, it raises
     :doc:`uproot.interpretation.identify.UnknownInterpretation`.
     """
+    # Match custom interpretations
+    matched_custom_interpretations = [
+        cls
+        for cls in _registered_interpretations
+        if cls.match_branch(branch, context, simplify)
+    ]
+
+    assert (
+        len(matched_custom_interpretations) <= 1
+    ), "Multiple custom interpretations matched, "
+    f"uproot cannot determine which one to use: {matched_custom_interpretations}"
+
+    if len(matched_custom_interpretations) == 1:
+        return matched_custom_interpretations[0](branch, context, simplify)
+
+    # Original interpretation logic
     if len(branch.branches) != 0:
         if branch.top_level and branch.has_member("fClassName"):
             typename = str(branch.member("fClassName"))
@@ -358,7 +405,7 @@ def interpretation_of(branch, context, simplify=True):
             uproot.containers.AsDynamic(), branch
         )
 
-    dims, is_jagged = _from_leaves(branch, context)
+    dims, _is_jagged = _from_leaves(branch, context)
 
     try:
         if len(branch.member("fLeaves")) == 0:
@@ -556,7 +603,7 @@ def _parse_ignore_extra_arguments(tokens, i, typename, file, at_most):
     while tokens[i].group(0) == ",":
         if at_most == 0:
             _parse_error(tokens[i].start() + 1, typename, file)
-        i, values = _parse_node(tokens, i + 1, typename, file, True, False, False)
+        i, _values = _parse_node(tokens, i + 1, typename, file, True, False, False)
         at_most -= 1
 
     return i
