@@ -25,7 +25,7 @@ import numpy
 import uproot.compression
 import uproot.extras
 import uproot.pyroot
-import uproot.writing._cascadetree
+import uproot.writing
 
 # To keep track of whether we've warned about switching to writing RNTuple by default
 _warned_rntuple_by_default = False
@@ -54,102 +54,11 @@ def add_to_directory(obj, name, directory, streamers):
     """
     is_ttree = False
 
-    if uproot._util.from_module(obj, "pandas"):
-        import pandas
-
-        if isinstance(
-            obj, pandas.DataFrame
-        ) and uproot._util.pandas_has_attr_is_numeric(pandas)(obj.index):
-            obj = uproot.writing._cascadetree.dataframe_to_dict(obj)
-
-    if uproot._util.from_module(obj, "awkward"):
-        import awkward
-
-        if isinstance(obj, awkward.Array):
-            obj = {"": obj}
-
-    if isinstance(obj, numpy.ndarray) and obj.dtype.fields is not None:
-        obj = uproot.writing._cascadetree.recarray_to_dict(obj)
+    obj = uproot.writing.writable._regularize_input_type(obj)
 
     if isinstance(obj, Mapping) and all(isinstance(x, str) for x in obj):
-        data = {}
-        metadata = {}
-
-        for branch_name, branch_array in obj.items():
-            if uproot._util.from_module(branch_array, "pandas"):
-                import pandas
-
-                if isinstance(branch_array, pandas.DataFrame):
-                    branch_array = uproot.writing._cascadetree.dataframe_to_dict(  # noqa: PLW2901 (overwriting branch_array)
-                        branch_array
-                    )
-
-            if (
-                isinstance(branch_array, numpy.ndarray)
-                and branch_array.dtype.fields is not None
-            ):
-                branch_array = uproot.writing._cascadetree.recarray_to_dict(  # noqa: PLW2901 (overwriting branch_array)
-                    branch_array
-                )
-
-            if isinstance(branch_array, Mapping) and all(
-                isinstance(x, str) for x in branch_array
-            ):
-                datum = {}
-                metadatum = {}
-                for kk, vv in branch_array.items():
-                    try:
-                        vv = (  # noqa: PLW2901 (overwriting vv)
-                            uproot._util.ensure_numpy(vv)
-                        )
-                    except TypeError:
-                        raise TypeError(
-                            f"unrecognizable array type {type(branch_array)} associated with {branch_name!r}"
-                        ) from None
-                    datum[kk] = vv
-                    branch_dtype = vv.dtype
-                    branch_shape = vv.shape[1:]
-                    if branch_shape != ():
-                        branch_dtype = numpy.dtype((branch_dtype, branch_shape))
-                    metadatum[kk] = branch_dtype
-
-                data[branch_name] = datum
-                metadata[branch_name] = metadatum
-
-            else:
-                if uproot._util.from_module(branch_array, "awkward"):
-                    data[branch_name] = branch_array
-                    metadata[branch_name] = branch_array.type
-
-                else:
-                    try:
-                        branch_array = uproot._util.ensure_numpy(  # noqa: PLW2901 (overwriting branch_array)
-                            branch_array
-                        )
-                    except TypeError:
-                        awkward = uproot.extras.awkward()
-                        try:
-                            branch_array = awkward.from_iter(  # noqa: PLW2901 (overwriting branch_array)
-                                branch_array
-                            )
-                        except Exception:
-                            raise TypeError(
-                                f"unrecognizable array type {type(branch_array)} associated with {branch_name!r}"
-                            ) from None
-                        else:
-                            data[branch_name] = branch_array
-                            metadata[branch_name] = awkward.type(branch_array)
-
-                    else:
-                        data[branch_name] = branch_array
-                        branch_dtype = branch_array.dtype
-                        branch_shape = branch_array.shape[1:]
-                        if branch_shape != ():
-                            branch_dtype = numpy.dtype((branch_dtype, branch_shape))
-                        metadata[branch_name] = branch_dtype
-
-        else:
-            is_ttree = True
+        is_ttree = True
+        metadata, data = uproot.writing.writable._unpack_metadata_and_arrays(obj)
 
     if is_ttree:
         global _warned_rntuple_by_default  # noqa: PLW0603
@@ -157,6 +66,7 @@ def add_to_directory(obj, name, directory, streamers):
             warnings.warn(
                 "Starting in version 5.7.0, Uproot will default to writing RNTuples instead of TTrees. "
                 "You will need to use `mktree` to explicitly create a TTree. "
+                "This can be done by changing `file['tree_name'] = data` to `file.mktree('tree_name', data)`. "
                 "Please update your code accordingly.",
                 FutureWarning,
                 stacklevel=4,
